package lcsb.mapviewer.converter.model.celldesigner.species;

import org.apache.log4j.Logger;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.exception.InvalidXmlSchemaException;
import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.converter.annotation.XmlAnnotationParser;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerElementCollection;
import lcsb.mapviewer.converter.model.celldesigner.annotation.RestAnnotationParser;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerAntisenseRna;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerCompartment;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerComplexSpecies;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerElement;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerGene;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerProtein;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerRna;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerSimpleMolecule;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerSpecies;
import lcsb.mapviewer.converter.model.celldesigner.structure.fields.CellDesignerModificationResidue;
import lcsb.mapviewer.converter.model.celldesigner.structure.fields.SpeciesState;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.compartment.PathwayCompartment;
import lcsb.mapviewer.model.map.kinetics.SbmlUnitType;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.species.AntisenseRna;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.map.species.field.ModificationState;
import lcsb.mapviewer.model.map.species.field.PositionToCompartment;

/**
 * Parser for sbml part of CellDesigner format. It's used for retrieving Species
 * from SBML.
 * 
 * @author Piotr Gawron
 * 
 */
public class SpeciesSbmlParser extends AbstractElementXmlParser<CellDesignerSpecies<?>, Species> {

  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private Logger logger = Logger.getLogger(SpeciesSbmlParser.class.getName());

  /**
   * Collection of {@link CellDesignerElement cell designer elements} parsed from
   * xml.
   */
  private CellDesignerElementCollection elements;

  /**
   * Default constructor. Model is required because some nodes require access to
   * other parts of the model.
   * 
   * @param elements
   *          collection of {@link CellDesignerElement cell designer elements}
   *          parsed from xml
   */
  public SpeciesSbmlParser(CellDesignerElementCollection elements) {
    this.elements = elements;
  }

  @Override
  public Pair<String, CellDesignerSpecies<?>> parseXmlElement(Node spieciesNode) throws InvalidXmlSchemaException {

    NodeList list = spieciesNode.getChildNodes();

    Node annotationNode = null;
    Node notesNode = null;
    for (int i = 0; i < list.getLength(); i++) {
      Node node = list.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        if (node.getNodeName().equals("annotation")) {
          annotationNode = node;
        } else if (node.getNodeName().equals("notes")) {
          notesNode = node;
        } else {
          throw new InvalidXmlSchemaException("Unknown element of annotation " + node.getNodeName());
        }
      }
    }

    if (annotationNode == null) {
      throw new InvalidXmlSchemaException("No annotation node in SBML/model/listOfSpecies/species");
    }

    list = annotationNode.getChildNodes();
    Node extensionNode = null;
    Node rdfNode = null;

    for (int i = 0; i < list.getLength(); i++) {
      Node node = list.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        if (node.getNodeName().equals("celldesigner:extension")) {
          extensionNode = node;
        } else if (node.getNodeName().equals("rdf:RDF")) {
          rdfNode = node;
        } else {
          throw new InvalidXmlSchemaException("Unknown element of annotation " + node.getNodeName());
        }
      }
    }

    if (extensionNode == null) {
      throw new InvalidXmlSchemaException(
          "No celldesigner:extension node in SBML/model/listOfSpecies/species/annotation");
    }

    list = extensionNode.getChildNodes();
    Node speciesIdentity = null;
    Node positionToCompartment = null;
    for (int i = 0; i < list.getLength(); i++) {
      Node node = list.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        if (node.getNodeName().equals("celldesigner:speciesIdentity")) {
          speciesIdentity = node;
        } else if (node.getNodeName().equals("celldesigner:positionToCompartment")) {
          positionToCompartment = node;
        } else if (node.getNodeName().equals("celldesigner:listOfCatalyzedReactions")) {
          // we can ignore it - every node on the map contains information about
          // itself
          continue;
        } else {
          throw new InvalidXmlSchemaException("[" + getNodeAttr("id", spieciesNode)
              + "]\tUnknown element of celldesigner:extension " + node.getNodeName());
        }
      }
    }

    if (speciesIdentity == null) {
      throw new InvalidXmlSchemaException(
          "No celldesigner:speciesIdentity node in SBML/model/listOfSpecies/species/annotation/extension");
    }

    Pair<String, CellDesignerSpecies<?>> result = parseSpeciesIdentity(speciesIdentity);

    CellDesignerSpecies<?> species = result.getRight();
    if (positionToCompartment != null) {
      PositionToCompartment position = PositionToCompartment.getByString(getNodeValue(positionToCompartment));
      if (position == null) {
        throw new InvalidXmlSchemaException("Unknown position on compartment: " + getNodeValue(positionToCompartment));
      }
      species.setPositionToCompartment(position);
    }

    // we ignore metaid - it's useless and obstruct data model
    // species.setMetaId(getNodeAttr("metaid", spieciesNode));
    species.setElementId(getNodeAttr("id", spieciesNode));
    species.setName(decodeName(getNodeAttr("name", spieciesNode)));
    species.setInitialAmount(getNodeAttr("initialAmount", spieciesNode));
    String substanceUnits = getNodeAttr("substanceUnits", spieciesNode);
    if (substanceUnits != null && !substanceUnits.isEmpty()) {
      species.setSubstanceUnits(SbmlUnitType.valueOf(substanceUnits.toUpperCase()));
    }
    species.setInitialAmount(getNodeAttr("initialAmount", spieciesNode));
    species.setInitialConcentration(getNodeAttr("initialConcentration", spieciesNode));
    if (species.getInitialAmount() == null && species.getInitialConcentration() == null) {
      species.setInitialConcentration(0.0);
    }
    species.setCharge(getNodeAttr("charge", spieciesNode));
    species.setOnlySubstanceUnits(getNodeAttr("hasOnlySubstanceUnits", spieciesNode));
    species.setBoundaryCondition(getNodeAttr("boundaryCondition", spieciesNode));
    species.setConstant(getNodeAttr("constant", spieciesNode));

    CellDesignerCompartment compartment = elements.getElementByElementId(getNodeAttr("compartment", spieciesNode));
    if (compartment != null) {
      species.setParent(compartment);
    }

    if (notesNode != null) {
      species.setNotes(getRap().getNotes(notesNode));
    }

    if (rdfNode != null) {
      XmlAnnotationParser xmlAnnotationParser = new XmlAnnotationParser();
      species.addMiriamData(xmlAnnotationParser.parseRdfNode(annotationNode.getChildNodes()));
    }

    return result;
  }

  @Override
  public String toXml(Species species) {
    StringBuilder builder = new StringBuilder();
    StringBuilder attributesBuilder = new StringBuilder();
    attributesBuilder.append(" name=\"" + escapeXml(encodeName(species.getName())) + "\"");
    attributesBuilder.append(" id=\"" + elements.getElementId(species) + "\"");
    attributesBuilder.append(" metaid=\"" + elements.getElementId(species) + "\"");
    if (species.getInitialAmount() != null) {
      attributesBuilder.append(" initialAmount=\"" + species.getInitialAmount() + "\"");
    }
    if (species.getInitialConcentration() != null) {
      attributesBuilder.append(" initialConcentration=\"" + species.getInitialConcentration() + "\"");
    }
    if (species.getCharge() != null) {
      attributesBuilder.append(" charge=\"" + species.getCharge() + "\"");
    }
    if (species.hasOnlySubstanceUnits() != null) {
      attributesBuilder.append(" hasOnlySubstanceUnits=\"" + species.hasOnlySubstanceUnits() + "\"");
    }
    if (species.getSubstanceUnits() != null) {
      attributesBuilder.append(" substanceUnits=\"" + species.getSubstanceUnits().toString().toLowerCase() + "\"");
    }
    if (species.getConstant() != null) {
      attributesBuilder.append(" constant=\"" + species.getConstant().toString().toLowerCase() + "\"");
    }
    if (species.getBoundaryCondition() != null) {
      attributesBuilder
          .append(" boundaryCondition=\"" + species.getBoundaryCondition().toString().toLowerCase() + "\"");
    }

    Compartment comp = null;
    // we have to exclude artificial compartment aliases, because they aren't
    // exported to CellDesigner file
    if (species.getCompartment() != null && !(species.getCompartment() instanceof PathwayCompartment)) {
      comp = species.getCompartment();
    } else if (species.getComplex() == null) {
      ModelData modelData = species.getModelData();
      if (modelData != null) {
        for (Compartment cAlias : modelData.getModel().getCompartments()) {
          if (!(cAlias instanceof PathwayCompartment) && cAlias.cross(species)) {
            if (comp == null) {
              comp = cAlias;
            } else if (comp.getSize() > cAlias.getSize()) {
              comp = cAlias;
            }
          }
        }
      }
    }

    if (comp != null) {
      attributesBuilder.append(" compartment=\"" + elements.getElementId(comp) + "\"");
    } else {
      attributesBuilder.append(" compartment=\"default\"");
    }
    builder.append("<species " + attributesBuilder.toString() + ">");

    if ((species.getNotes() != null && !species.getNotes().equals("")) || species.getSymbol() != null
        || species.getFullName() != null || species.getSynonyms().size() > 0) {
      builder.append("<notes>");
      builder.append("<html xmlns=\"http://www.w3.org/1999/xhtml\"><head><title/></head><body>");
      RestAnnotationParser rap = new RestAnnotationParser();
      builder.append(rap.createAnnotationString(species));
      if (species.getNotes() != null) {
        builder.append(species.getNotes());
      }
      builder.append("</body></html>");
      builder.append("</notes>\n");
    }
    builder.append("<annotation>");
    builder.append("<celldesigner:extension>");
    if (species.getPositionToCompartment() != null) {
      builder.append("<celldesigner:positionToCompartment>" + species.getPositionToCompartment().getStringName()
          + "</celldesigner:positionToCompartment>");
    }
    builder.append(speciesIdentityToXml(species));

    builder.append("</celldesigner:extension>\n");

    XmlAnnotationParser xmlAnnotationParser = new XmlAnnotationParser();
    builder.append(xmlAnnotationParser.dataSetToXmlString(species.getMiriamData()));

    builder.append("</annotation>\n");

    builder.append("</species>\n");
    return builder.toString();
  }

  /**
   * Creates speciesIdentity xml node for the {@link CellDesignerSpecies}.
   * 
   * @param species
   *          species to be transformed into xml
   * @return speciesIdentity xml node
   */
  String speciesIdentityToXml(Species species) {
    StringBuilder sb = new StringBuilder();
    String classType = "";
    SpeciesMapping mapping = SpeciesMapping.getMappingByModelClass(species.getClass());
    if (mapping != null) {
      classType = mapping.getCellDesignerString();
    } else {
      throw new InvalidArgumentException("Invalid species class type: " + species.getClass().getName());
    }
    sb.append("<celldesigner:speciesIdentity>");
    sb.append("<celldesigner:class>" + classType + "</celldesigner:class>\n");
    if (species instanceof Rna) {
      sb.append("<celldesigner:rnaReference>r_" + elements.getElementId(species) + "</celldesigner:rnaReference>\n");
    } else if (species instanceof Protein) {
      sb.append(
          "<celldesigner:proteinReference>p_" + elements.getElementId(species) + "</celldesigner:proteinReference>\n");
    } else if (species instanceof Gene) {
      sb.append("<celldesigner:geneReference>g_" + elements.getElementId(species) + "</celldesigner:geneReference>\n");
    } else if (species instanceof AntisenseRna) {
      sb.append("<celldesigner:antisensernaReference>ar_" + elements.getElementId(species)
          + "</celldesigner:antisensernaReference>\n");
    }
    SpeciesState state = new SpeciesState(species);
    sb.append(speciesStateToXml(state));
    sb.append("<celldesigner:name>" + escapeXml(encodeName(species.getName())) + "</celldesigner:name>\n");

    Boolean hypothetical = species.getHypothetical();
    if (hypothetical != null) {
      sb.append("<celldesigner:hypothetical>" + hypothetical + "</celldesigner:hypothetical>\n");
    }

    sb.append("</celldesigner:speciesIdentity>\n");
    return sb.toString();
  }

  /**
   * Creates CellDesigner species state into xml node.
   * 
   * @param state
   *          state of the species.
   * @return CellDesigner xml node used for description of the state
   */
  String speciesStateToXml(SpeciesState state) {
    StringBuilder sb = new StringBuilder();
    if (state.getHomodimer() != 1) {
      sb.append("<celldesigner:homodimer>" + state.getHomodimer() + "</celldesigner:homodimer>\n");
    }
    if (state.getStructuralState() != null && !state.getStructuralState().equals("")) {
      sb.append("<celldesigner:listOfStructuralStates>\n");
      sb.append("<celldesigner:structuralState structuralState=\"" + state.getStructuralState() + "\">"
          + state.getStructuralState() + "</celldesigner:structuralState>");
      sb.append("</celldesigner:listOfStructuralStates>\n");
    }
    if (state.getModifications().size() > 0) {
      sb.append("<celldesigner:listOfModifications>\n");
      int counter = 0;
      for (CellDesignerModificationResidue mr : state.getModifications()) {
        sb.append(modificationResidueToXml(mr, counter++));
      }
      sb.append("</celldesigner:listOfModifications>\n");
    }
    if (sb.length() > 0) {
      sb.insert(0, "<celldesigner:state>\n");
      sb.append("</celldesigner:state>\n");
    }
    return sb.toString();
  }

  /**
   * Parses CellDesigner SpeciesIdentity node into {@link CellDesignerSpecies}.
   * 
   * @param rootNode
   *          xml node to parse
   * @return {@link Pair} of Species retrieved from SpeciesIdentity node and
   *         CellDesigner identifier
   * @throws InvalidXmlSchemaException
   *           thrown when there is a problem with xml
   */
  public Pair<String, CellDesignerSpecies<?>> parseSpeciesIdentity(Node rootNode) throws InvalidXmlSchemaException {
    NodeList nodes = rootNode.getChildNodes();
    CellDesignerSpecies<?> result = new CellDesignerSpecies<Species>();
    Node classNode = getNode("celldesigner:class", nodes);
    if (classNode == null) {
      throw new InvalidXmlSchemaException("Species node in Sbml model doesn't contain node \"celldesigner:class\".");
    } else {
      String value = getNodeValue(classNode);
      SpeciesMapping mapping = SpeciesMapping.getMappingByString(value);
      if (mapping != null) {
        result = mapping.createSpecies(result);
      } else {
        throw new InvalidXmlSchemaException("Species node in Sbml model is of unknown type: " + value);
      }
    }
    String identifier = "";
    for (int x = 0; x < nodes.getLength(); x++) {
      Node node = nodes.item(x);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        if (node.getNodeName().equalsIgnoreCase("celldesigner:class")) {
          continue;
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:proteinReference")) {
          if (result instanceof CellDesignerProtein) {
            identifier = getNodeValue(node);
          } else {
            throw new InvalidXmlSchemaException(
                "Wrong class type for protein reference: " + result.getClass().getName());
          }
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:geneReference")) {
          if (result instanceof CellDesignerGene) {
            identifier = getNodeValue(node);
          } else {
            throw new InvalidXmlSchemaException("Wrong class type for gene reference: " + result.getClass().getName());
          }
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:rnaReference")) {
          if (result instanceof CellDesignerRna) {
            identifier = getNodeValue(node);
          } else {
            throw new InvalidXmlSchemaException("Wrong class type for rna reference: " + result.getClass().getName());
          }
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:antisensernaReference")) {
          if (result instanceof CellDesignerAntisenseRna) {
            identifier = getNodeValue(node);
          } else {
            throw new InvalidXmlSchemaException(
                "Wrong class type for antisense rna reference: " + result.getClass().getName());
          }
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:hypothetical")) {
          result.setHypothetical(getNodeValue(node));
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:name")) {
          result.setName(decodeName(getNodeValue(node)));
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:state")) {
          SpeciesState state = parseXmlSpeciesState(node);
          processStateDataInSpecies(result, state);
        } else {
          throw new InvalidXmlSchemaException("Unknown element of celldesigner:speciesIdentity: " + node.getNodeName());
        }
      }

    }
    return new Pair<String, CellDesignerSpecies<?>>(identifier, result);
  }

  /**
   * Set structural state to species. StructuralState is a generic class with some
   * information about the species element.
   * 
   * @param state
   *          state to be set
   * 
   * @param species
   *          where the data should be saved
   */
  void processStateDataInSpecies(CellDesignerSpecies<?> species, SpeciesState state) {
    species.setHomodimer(state.getHomodimer());

    if (species instanceof CellDesignerComplexSpecies) {
      CellDesignerComplexSpecies complex = (CellDesignerComplexSpecies) species;
      complex.setStructuralState(state.getStructuralState());
      if (state.getModifications().size() != 0) {
        throw new NotImplementedException("Modification not supported in Complex");
      }
    } else if (species instanceof CellDesignerGene) {
      CellDesignerGene gene = (CellDesignerGene) species;
      if (state.getStructuralState() != null) {
        throw new NotImplementedException("StructuralState not supported in Gene");
      }
      for (CellDesignerModificationResidue mr : state.getModifications()) {
        gene.addModificationResidue(mr);
      }
    } else if (species instanceof CellDesignerProtein) {
      CellDesignerProtein<?> protein = (CellDesignerProtein<?>) species;
      protein.setStructuralState(state.getStructuralState());
      for (CellDesignerModificationResidue mr : state.getModifications()) {
        protein.addModificationResidue(mr);
      }
    } else if (species instanceof CellDesignerRna) {
      CellDesignerRna rna = (CellDesignerRna) species;
      if (state.getStructuralState() != null && !state.getStructuralState().isEmpty()) {
        throw new NotImplementedException("Structural state not supported in RNA");
      }

      for (CellDesignerModificationResidue mr : state.getModifications()) {
        rna.addRegion(mr);
      }
    } else if (species instanceof CellDesignerSimpleMolecule) {
      if (state.getStructuralState() != null && !state.getStructuralState().isEmpty()) {
        throw new NotImplementedException("Structural state not supported in SimpleMolecule");
      }
      if (state.getModifications().size() != 0) {
        throw new NotImplementedException("Modification not supported in SimpleMolecule");
      }

    } else if (species instanceof CellDesignerAntisenseRna) {
      CellDesignerAntisenseRna antisenseRna = (CellDesignerAntisenseRna) species;
      if (state.getStructuralState() != null && !state.getStructuralState().isEmpty()) {
        throw new NotImplementedException("Structural state not supported in AntisenseRna");
      }

      for (CellDesignerModificationResidue mr : state.getModifications()) {
        antisenseRna.addRegion(mr);
      }
    } else {
      if (state.getStructuralState() != null && !state.getStructuralState().isEmpty()) {
        throw new NotImplementedException("Structural state not supported in " + species.getClass().getSimpleName());
      }
      if (state.getModifications().size() != 0) {
        throw new NotImplementedException("Modification not supported in " + species.getClass().getSimpleName());
      }
    }

  }

  /**
   * Creates {@link SpeciesState} from xml node.
   * 
   * @param stateNode
   *          xml node
   * @return {@link SpeciesState} retrieved from xml node
   * @throws InvalidXmlSchemaException
   *           thrown when xml node contains data that is not supported by xml
   *           schema
   */
  private SpeciesState parseXmlSpeciesState(Node stateNode) throws InvalidXmlSchemaException {
    NodeList nodes = stateNode.getChildNodes();
    SpeciesState result = new SpeciesState();
    for (int x = 0; x < nodes.getLength(); x++) {
      Node node = nodes.item(x);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        if (node.getNodeName().equalsIgnoreCase("celldesigner:homodimer")) {
          result.setHomodimer(getNodeValue(node));
        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:listOfStructuralStates")) {
          NodeList structureNodeList = node.getChildNodes();
          for (int y = 0; y < structureNodeList.getLength(); y++) {
            Node structureNode = structureNodeList.item(y);
            if (structureNode.getNodeType() == Node.ELEMENT_NODE) {
              if (structureNode.getNodeName().equalsIgnoreCase("celldesigner:structuralState")) {
                String state = getNodeAttr("structuralState", structureNode);
                result.setStructuralState(state);
              } else {
                throw new InvalidXmlSchemaException(
                    "Unknown element of celldesigner:listOfStructuralStates: " + node.getNodeName());
              }

            }

          }

        } else if (node.getNodeName().equalsIgnoreCase("celldesigner:listOfModifications")) {
          NodeList modificationNodeList = node.getChildNodes();
          for (int y = 0; y < modificationNodeList.getLength(); y++) {
            Node structureNode = modificationNodeList.item(y);
            if (structureNode.getNodeType() == Node.ELEMENT_NODE) {
              if (structureNode.getNodeName().equalsIgnoreCase("celldesigner:modification")) {
                CellDesignerModificationResidue modification = getSpeciesModification(structureNode);

                result.addModificationResidue(modification);
              } else {
                throw new InvalidXmlSchemaException(
                    "Unknown element of celldesigner:listOfModifications: " + node.getNodeName());
              }
            }
          }
        } else {
          throw new InvalidXmlSchemaException("Unknown element of celldesigner:state: " + node.getNodeName());
        }
      }
    }
    return result;
  }

  /**
   * Creates {@link CellDesignerModificationResidue} from the appropriate xml
   * node.
   * 
   * @param rootNode
   *          xml node
   * @return {@link CellDesignerModificationResidue} retrieved from the node
   * @throws InvalidXmlSchemaException
   *           thrown when xml node contains data that is not supported by xml
   *           schema
   */
  private CellDesignerModificationResidue getSpeciesModification(Node rootNode) throws InvalidXmlSchemaException {
    CellDesignerModificationResidue modification = new CellDesignerModificationResidue();
    modification.setIdModificationResidue(getNodeAttr("residue", rootNode));
    String state = getNodeAttr("state", rootNode);
    if (state != null) {
      ModificationState mState = ModificationState.getByName(state);
      if (mState == null) {
        throw new InvalidXmlSchemaException("[" + modification.getClass().getSimpleName() + "\t"
            + modification.getIdModificationResidue() + "] Unknown modification state: " + state);
      }
      modification.setState(mState);
    }
    modification.setName(getNodeAttr("name", rootNode));
    modification.setSize(getNodeAttr("size", rootNode));
    modification.setAngle(getNodeAttr("pos", rootNode));
    NodeList list = rootNode.getChildNodes();
    for (int y = 0; y < list.getLength(); y++) {
      Node node = list.item(y);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        throw new InvalidXmlSchemaException(
            "[" + modification.getClass().getSimpleName() + "\t" + modification.getIdModificationResidue()
                + "] Unknown element of celldesigner:modification: " + node.getNodeName());
      }
    }
    return modification;
  }

  /**
   * Creates xml node representing {@link CellDesignerModificationResidue}.
   * 
   * @param mr
   *          object to be transformed into xml
   * @return xml node representing param object
   */
  private String modificationResidueToXml(CellDesignerModificationResidue mr, int number) {
    String state = "";
    if (mr.getState() != null) {
      state = mr.getState().getFullName();
    }
    if (state == null || state.equals("")) {
      return "";
    }
    return "<celldesigner:modification residue=\"" + elements.getModificationResidueId(mr, number)
        + "\" state=\"" + state + "\"> </celldesigner:modification>\n";
  }
}
