package lcsb.mapviewer.api.convert;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLStreamException;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.sbml.jsbml.SBMLException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.converter.ConverterException;
import lcsb.mapviewer.converter.ConverterParams;
import lcsb.mapviewer.converter.IConverter;
import lcsb.mapviewer.converter.InvalidInputDataExecption;
import lcsb.mapviewer.converter.graphics.AbstractImageGenerator;
import lcsb.mapviewer.converter.graphics.DrawingException;
import lcsb.mapviewer.converter.graphics.ImageGenerators;
import lcsb.mapviewer.model.map.InconsistentModelException;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.services.SecurityException;

@Transactional
@Service
public class ConvertRestImpl extends BaseRestImpl {
	
	  @SuppressWarnings("unused")
	  private Logger logger = Logger.getLogger(ConvertRestImpl.class);


  public String convert(String token, String fromFormat, String toFormat, String input)
		  throws SecurityException, InvalidInputDataExecption, 
	      SBMLException, InconsistentModelException, IOException, ConverterException, 
	      XMLStreamException, DrawingException, QueryException 
  {
	  ConverterParams params = createConvertParams(input);	  
	  Model model = getModelParserByNameOrClass(fromFormat).createModel(params);
	  
	  IConverter exporter =  getModelParserByNameOrClass(toFormat);
	  return IOUtils.toString(exporter.exportModelToInputStream(model));
  }
  
  public ByteArrayOutputStream converToImage(String token, String fromFormat, String toFormat, String input,
		  Double targetWidth, Double targetHeight) 
		  throws SecurityException, InvalidInputDataExecption,
		  SBMLException, InconsistentModelException, IOException, 
		  ConverterException,XMLStreamException, DrawingException, 
		  QueryException 
  {
	  Model model = getModelParserByNameOrClass(fromFormat).createModel(createConvertParams(input));
		  
	  AbstractImageGenerator generator = getImageGenerator(toFormat, createImageParams(model, targetWidth, targetHeight));	
	  ByteArrayOutputStream os = new ByteArrayOutputStream();
	  generator.saveToOutputStream(os);
	  return os;
  }
  
  public ByteArrayOutputStream converToImage(String token, String fromFormat, String toFormat, String input) 
		  throws SecurityException, InvalidInputDataExecption,
		  SBMLException, InconsistentModelException, IOException, 
		  ConverterException,XMLStreamException, DrawingException, 
		  QueryException 
  {
	  return converToImage(token, fromFormat, toFormat, input, 0.0, 0.0);
  }
  
  public Map<String, Object> getInformation(String token){
	  Map<String, Object> info =  new LinkedHashMap<>();
	  
	  	  
	  List<Object> converters = new ArrayList<>();
	  for (IConverter converter: getModelConverters()) {
		  
		  List<String> names = new ArrayList<>();		  
		  names.add(removeWhiteSpaces(converter.getCommonName()));
		  names.add(converter.getClass().getCanonicalName());
		  
		  Map<String, Object> names_item = new LinkedHashMap<>();
		  names_item.put("available_names", names);
		  converters.add(names_item);
	  }
	  
	  info.put("inputs", converters);
	  info.put("outputs", converters);
	  return info;	  
  }
  
  public Map<String, Object> getInformationImage(String token){
	  Map<String, Object> info =  getInformation(token);
	  
	  List<Object> generators = new ArrayList<>();	  
	  ImageGenerators igs = new ImageGenerators();	  
	  for (Pair<String, Class<? extends AbstractImageGenerator>> generator: igs.getAvailableImageGenerators()) {
		  
		  List<String> names = new ArrayList<>();
		  names.add(igs.getExtension(generator.getRight()));
		  names.add(generator.getRight().getCanonicalName());
		  
		  Map<String, Object> names_item = new LinkedHashMap<>();
		  names_item.put("available_names", names);
		  generators.add(names_item);
	  }
	  
	  info.remove("outputs");	  
	  info.put("outputs", generators);
	  return info;	  
  }
  
  private IConverter getModelParserByNameOrClass(String id) throws QueryException {
	  try {
		  return getModelParserByName(id);
	  } catch (QueryException e) {
		  return getModelParser(id);
	  }	  
  }
  
  private String removeWhiteSpaces(String str) {
	  return str.replace(' ', '_');
  }
  
  private IConverter getModelParserByName(String name) throws QueryException {
	  for (IConverter converter : getModelConverters()) {
		  if (removeWhiteSpaces(converter.getCommonName()).equals(name)) {
			  return converter;
		  }
	  }
	  throw new QueryException("Unknown parser name: " + name);
  }
  
  private AbstractImageGenerator getImageGenerator(String extOrClass, AbstractImageGenerator.Params params) throws QueryException
  {	  
	  for (Pair<String, Class<? extends AbstractImageGenerator>> element : new ImageGenerators().getAvailableImageGenerators()) {		  	
		  	
			try {
				Class<? extends AbstractImageGenerator> clazz = element.getRight();				
				Constructor<?> ctor = clazz.getConstructor(AbstractImageGenerator.Params.class);
				AbstractImageGenerator generator = (AbstractImageGenerator) ctor.newInstance(params);				
				if (extOrClass.equals(clazz.getCanonicalName()) || extOrClass.equals(generator.getFileExtension())) {
					return generator;
				}
				
			} catch (NoSuchMethodException | java.lang.SecurityException | InstantiationException 
					| IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
				logger.error("Creation of image generator class for '" + element.getLeft() + "' failed.");
				throw new QueryException("Issue with obtaining image generator for extension " + extOrClass + ".");
			}
		  	
		}
	  
	  throw new QueryException("Image generator for extension " + extOrClass + " not available.");
  } 
  
  private AbstractImageGenerator.Params createImageParams(Model model, Double targetWidth, Double targetHeight){  
	  
	  
	  //if (targetHeight == 0) targetHeight = 848.0;
	  //if (targetWidth == 0) targetWidth = 595.0;
	  
	  Double padding = 5.0;
	  Double maxDim = 10000.0;
	  
	  Double w = model.getWidth();
	  Double h = model.getHeight();
	  
	  if (targetHeight == 0) targetHeight = Math.min(h, maxDim);
	  if (targetWidth == 0) targetWidth = Math.min(w, maxDim);
	  
	  Double scale = targetWidth / w;  
	  if (h * scale > targetHeight) {
		  scale = targetHeight / h;		  
	  }
	  
	  Double wScaled = w*scale;
	  Double hScaled = h*scale;
	  
	  return new AbstractImageGenerator.Params().
				model(model).
				width(wScaled + padding).
				height(hScaled + padding).
				scale(1/scale)
				.x(0);
				
  }  

private ConverterParams createConvertParams(String input) {
	  ConverterParams params = new ConverterParams();
	  InputStream is = new ByteArrayInputStream(input.getBytes());
	  
	  params.inputStream(is);
	  
	  return params;
  }
}
