package lcsb.mapviewer.api.projects.models;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.nullable;
import static org.mockito.ArgumentMatchers.refEq;

import java.util.List;
import java.util.Map;
import java.util.Set;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;

import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.api.RestTestFunctions;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.converter.graphics.PdfImageGenerator;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.cache.FileEntry;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IProjectService;

public class ModelRestImplTest extends RestTestFunctions {

  @Autowired
  ModelRestImpl _modelRestImpl;

  @Autowired
  IModelService modelService;

  @Autowired
  IProjectService projectService;

  @Before
  public void setUp() throws Exception {
    _modelRestImpl.setModelService(modelService);
    _modelRestImpl.setProjectService(projectService);
  }

  @After
  public void tearDown() throws Exception {
  }

  @Test
  public void testModels() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      List<Map<String, Object>> result = modelRest.getModels("sample", token);
      assertEquals(1, result.size());
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testGetEmptyModels() throws Exception {
    try {
      ModelRestImpl modelRest = createEmptyProjectRest("test");
      List<Map<String, Object>> result = modelRest.getModels("test", token);
      assertEquals(0, result.size());
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  private ModelRestImpl createEmptyProjectRest(String string) throws Exception {
    IModelService mockModelService = Mockito.mock(IModelService.class);
    Mockito.when(mockModelService.getLastModelByProjectId(refEq(string), any())).thenReturn(null);

    IProjectService mockProjectService = Mockito.mock(IProjectService.class);
    Mockito.when(mockProjectService.getProjectByProjectId(refEq(string), any())).thenReturn(new Project());

    _modelRestImpl.setModelService(mockModelService);
    _modelRestImpl.setProjectService(mockProjectService);
    return _modelRestImpl;
  }

  private ModelRestImpl createMockProjectRest(String string) throws Exception {
    Model model = super.getModelForFile(string, true);
    Project project = new Project();
    project.addModel(model);

    IModelService mockModelService = Mockito.mock(IModelService.class);
    Mockito.when(mockModelService.getLastModelByProjectId(nullable(String.class), any())).thenReturn(model);
    _modelRestImpl.setModelService(mockModelService);

    IProjectService mockProjectService = Mockito.mock(IProjectService.class);
    Mockito.when(mockProjectService.getProjectByProjectId(nullable(String.class), any())).thenReturn(project);
    _modelRestImpl.setProjectService(mockProjectService);

    return _modelRestImpl;
  }

  @Test
  public void testGetModelAsImageForInvalidConverter() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      modelRest.getModelAsImage(token, "sample", "0", "", "", "", "", "");
      fail("Exception expected");
    } catch (InvalidArgumentException e) {
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testGetModelAsImage() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      FileEntry result = modelRest.getModelAsImage(token, "sample", "0", PdfImageGenerator.class.getCanonicalName(), "",
          "", "", "");
      assertNotNull(result);
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testGetModelAsImage2() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      modelRest.getModelAsImage(token, "sample", "0", PdfImageGenerator.class.getCanonicalName(), "", "", "", "");
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testGetModelAsFileModel() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      modelRest.getModelAsModelFile(token, "sample", "0", "", "", "", "", "", null, null);
      fail("Exception expected");
    } catch (QueryException e) {
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testGetModelAsFileModel2() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      modelRest.getModelAsModelFile(token, "sample", "0", CellDesignerXmlParser.class.getCanonicalName(), "", "", "",
          "0,0;90,0;90,90;90,0", null, null);
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testGetModelAsFileModel3() throws Exception {
    try {
      ModelRestImpl modelRest = createMockProjectRest("testFiles/model/sample.xml");
      modelRest.getModelAsModelFile(token, "sample", "0", "", "", "", "", "0,0;90,0;90,90;90,0", null, null);
      fail("Exception expected");
    } catch (QueryException e) {
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  @Test
  public void testEmptyStringListToIntegerList() throws Exception {
    try {
      Set<Integer> ids = _modelRestImpl.stringListToIntegerSet("");
      assertNotNull(ids);
      assertEquals(0, ids.size());
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }
  @Test
  public void testStringListToIntegerArray() throws Exception {
    try {
      Set<Integer> ids = _modelRestImpl.stringListToIntegerSet("12,3");
      assertNotNull(ids);
      assertEquals(2, ids.size());
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }
}
